#pragma once

#include "configuration.hh"
#include "input_pickers/critical.hh"
#include "input_pickers/critical_fullrnd.hh"
#include "input_pickers/critical_pq.hh"
#include "input_pickers/critical_rnd.hh"
